/*
 * Portions copyright 2002, 2003 Richard Laager.
 * Copyright (c) 1996, 1997, 1998, 1999, Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARC HOROWITZ ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL MARC HOROWITZ BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

const char rcsid_kd_delete_c[] =
  "$Id: kd_delete.c,v 1.9 2003/07/31 18:19:02 rlaager Exp $";

#if HAVE_CTYPE_H
#include <ctype.h>
#endif

#if HAVE_ERRNO_H
#include <errno.h>
#endif

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "database.h"
#include "globals.h"
#include "kd_internal.h"
#include "kd_search.h"
#include "kd_types.h"
#include "llist.h"
#include "pgputil.h"
#include "util.h"

extern int txn_count;
extern int use_txn;

typedef struct _dwfw_state
{
  unsigned char entry[12];
  error *err;
}
dwfw_state;

typedef struct _dkfw_state
{
  xbuffer *deleted;
}
dkfw_state;


/* this isn't particularly efficient, but deletes aren't all that common. */
int
delete_word_from_worddb (void *e, void *c)
{
  words_elem *we = (words_elem *) e;
  dwfw_state *s = (dwfw_state *) c;

  unsigned char word[256];
  DBC *cursor;
  DB_TXN *tid = NULL;
  DBT key, data;
  int i, ret;

  memset (&key, 0, sizeof (key));
  memset (&data, 0, sizeof (data));

  for (i = 0; i < we->len && i < sizeof (word); i++)
    word[i] = tolower ((we->ptr)[i]);

  key.data = (void *) word;
  key.size = (size_t) we->len;

  data.data = s->entry;
  data.size = 12;

  if (use_txn)
    worddb->dbenv->txn_begin (worddb->dbenv, NULL, &tid, 0);
  else
    tid = NULL;

  ret = worddb->cursor (worddb, tid, &cursor, 0);

  if (ret && (ret != DB_NOTFOUND))
    {
      if (use_txn)
	tid->abort (tid);
      s->err->fatal = 1;
      s->err->str = "database read error creating worddb cursor for delete";
      fail ();
    }

  if ((ret = (*(cursor->c_get)) (cursor, &key, &data, DB_GET_BOTH)))
    {
      /* We can't pass non-fatal error messages back, so we have to print
         them ourselves.  This should be OK since only pksclient does
         deletions. */
      (*(cursor->c_close)) (cursor);
      if (use_txn)
	tid->abort (tid);
      fflush (stdout);
      fprintf (stderr,
	       "warning:  consistency error reading worddb for delete: %.*s not found\n",
	       (int) we->len, we->ptr);
      return (1);
    }

  /* loop over the matching entries, deleting as we go.  Normally, there
     can be only one matching entry, but there was once a bug which
     would leave extra matches around */

  for (;
       ret == 0; ret = (*(cursor->c_get)) (cursor, &key, &data, DB_NEXT_DUP))
    {
      if (memcmp (data.data, s->entry, 12) != 0)
	break;

      if ((ret = (*(cursor->c_del)) (cursor, 0)))
	break;
    }

  (*(cursor->c_close)) (cursor);
  if (use_txn)
    TXN_COMMIT (tid, worddb);
  if (ret && (ret != DB_NOTFOUND))
    {
      s->err->fatal = 1;
      sprintf (s->err->buf,
	       "error reading worddb entry for delete (errno = %d)", ret);
      fail ();
    }

  return (1);
}

int
delete_key_from_worddb (void *e, llist * new_list, void *c, error * err)
{
  keys_elem *ke = (keys_elem *) e;
  dkfw_state *s = (dkfw_state *) c;
  dwfw_state dwfws;
  char buf[128];

  kd_make_worddb_entry (ke, dwfws.entry);
  dwfws.err = err;

  if (!llist_iterate (&(ke->words), delete_word_from_worddb, &dwfws))
    return (0);

  sprintf (buf, "key id %02X%02X%02X%02X deleted\n",
	   ke->keyidbits.buf[4],
	   ke->keyidbits.buf[5], ke->keyidbits.buf[6], ke->keyidbits.buf[7]);

  if (!xbuffer_append_str (s->deleted, buf))
    return (0);

  keys_elem_free (e, NULL);

  return (1);
}

int
kd_delete_1 (unsigned char *userid, long len, int flags,
	     xbuffer * deleted, error * err)
{
  dkfw_state dkfws;
  int ret;

  /* the NULL here as the append function means that any matching key
     will just disappear.  delete_key_from_worddb will remove the
     words as a side effect. */

  dkfws.deleted = deleted;

  ret = kd_search_1 (userid, len, KD_SEARCH_EXACT, -1,
		     NULL, delete_key_from_worddb, &dkfws, err);

  kd_sync ();

  if (!ret)
    return (0);

  if (deleted->len == 0)
    {
      /* no matching keys in database */

      err->fatal = 0;
      err->str = "No matching keys in database";
      return (0);
    }

  return (1);
}

int
kd_delete (unsigned char *userid, long len, int flags,
	   unsigned char **ret, long *retlen)
{
  error err;
  xbuffer deleted;

  err.str = err.buf;

  xbuffer_alloc (&deleted);

  kd_log_start ("kd_delete", userid, len, flags);

  if (kd_delete_1 (userid, len, flags, &deleted, &err))
    {
      *ret = deleted.buf;
      *retlen = deleted.len;

      kd_log_finish ("kd_delete", 1);

      return (1);
    }

  if (!err.fatal)
    {
      if (!(*ret = (unsigned char *) my_strdup (err.str)))
	{
	  err.fatal = 1;
	  err.str = "Failed allocating space for error string";
	  dabort ();

	  /* fall through to fatal error handler */
	}
      else
	{
	  *retlen = strlen ((char *) *ret);

	  kd_log_finish ("kd_delete", 0);

	  return (0);
	}
    }

  /* fatal errors */

  if (err.fatal)
    {
      log_fatal ("kd_delete", err.str);
      /* never returns */
    }

  /* keep the compiler quiet */

  return (0);
}
