/*
 * Portions copyright 2002, 2003 Richard Laager.
 * Copyright (c) 1996, 1997, 1998, 1999, Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARC HOROWITZ ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL MARC HOROWITZ BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

const char rcsid_kd_disable_c[] =
  "$Id: kd_disable.c,v 1.6 2003/02/28 17:17:13 rlaager Exp $";

#if HAVE_CTYPE_H
#include <ctype.h>
#endif

#if HAVE_ERRNO_H
#include <errno.h>
#endif

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "database.h"
#include "globals.h"
#include "kd_internal.h"
#include "kd_search.h"
#include "kd_types.h"
#include "llist.h"
#include "pgputil.h"

typedef struct _cd_state
{
  xbuffer *xb;
  int set;
}
cd_state;

int
change_disable (void *e, llist * new_list, void *c, error * err)
{
  keys_elem *ke = (keys_elem *) e;
  cd_state *s = (cd_state *) c;
  char buf[128];

  ke->disabled = (s->set) ? -1 : 0;

  if (!llist_add (new_list, e))
    {
      err->fatal = 1;
      err->str =
	(s->set
	 ? "appending disabled key to new_list failed"
	 : "appending undisabled key to new_list failed");
      fail ();
    }

  sprintf (buf, "key id %02X%02X%02X%02X %s\n",
	   ke->keyidbits.buf[4],
	   ke->keyidbits.buf[5],
	   ke->keyidbits.buf[6],
	   ke->keyidbits.buf[7], s->set ? "disabled" : "undisabled");

  if (!xbuffer_append_str (s->xb, buf))
    return (0);

  return (1);
}

int
kd_disable_1 (unsigned char *userid, long len, int flags,
	      xbuffer * disabled, error * err)
{
  int ret;
  cd_state cds;

  cds.xb = disabled;
  cds.set = (flags & KD_DISABLE_CLEAR) ? 0 : 1;

  ret = kd_search_1 (userid, len, KD_SEARCH_EXACT, -1,
		     NULL, change_disable, &cds, err);

  kd_sync ();

  if (!ret)
    return (0);

  if (disabled->len == 0)
    {
      /* no matching keys in database */

      err->fatal = 0;
      err->str = "No matching keys in database";
      return (0);
    }

  return (1);
}

int
kd_disable (unsigned char *userid, long len, int flags,
	    unsigned char **ret, long *retlen)
{
  error err;
  xbuffer disabled;

  err.str = err.buf;

  xbuffer_alloc (&disabled);

  kd_log_start ("kd_disable", userid, len, flags);

  if (kd_disable_1 (userid, len, flags, &disabled, &err))
    {
      *ret = disabled.buf;
      *retlen = disabled.len;

      kd_log_finish ("kd_disabled", 1);

      return (1);
    }

  if (!err.fatal)
    {
      if (!(*ret = (unsigned char *) my_strdup (err.str)))
	{
	  err.fatal = 1;
	  err.str = "Failed allocating space for error string";
	  dabort ();

	  /* fall through to fatal error handler */
	}
      else
	{
	  *retlen = strlen ((char *) *ret);

	  kd_log_finish ("kd_disable", 0);

	  return (0);
	}
    }

  /* fatal errors */

  if (err.fatal)
    {
      log_fatal ("kd_disable", err.str);
      /* never returns */
    }

  /* keep the compiler quiet */

  return (0);
}
