/*
 * Portions copyright 2002, 2003 Richard Laager.
 * Copyright (c) 1996, 1997, 1998, 1999, Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARC HOROWITZ ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL MARC HOROWITZ BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

const char rcsid_kd_generic_c[] =
  "$Id: kd_generic.c,v 1.12 2003/07/31 18:19:02 rlaager Exp $";

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#if HAVE_CTYPE_H
#include <ctype.h>
#endif

#if HAVE_ERRNO_H
#include <errno.h>
#endif

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "database.h"
#include "globals.h"
#include "llist.h"
#include "kd_types.h"
#include "kd_internal.h"

/* this file contains functions which are common to two different
   database operations, and functions which perform operations
   on the database itself. */

/* multiple keydb files are supported.  which file is used is
   determined by the lsb 16 bits of the keyid mod the number of
   databases.  The number of database files to use is specified at
   create-time, written out, and then learned at run-time by reading
   the file from the dbdir. */

static DB_ENV *dbenv;
extern int txn_count;
extern int use_txn;

int num_keydb;
DB **keydb_files = NULL;
DB *worddb = NULL;
DB *timedb = NULL;

int
get_keydb_num (DBT * key)
{
  /* keyid's are 8 bytes, msb first.  so start from the end.  use 16
     bits, since that's enough to divide by any small number of db
     files. */
  unsigned char *keydata = (unsigned char *) key->data;
  unsigned long keyidnum;

  keyidnum =
    (keydata[KEYDB_KEYID_BYTES - 2] << 8) | keydata[KEYDB_KEYID_BYTES - 1];

  return (keyidnum % num_keydb);
}

DB *
keydb (DBT * key)
{
  return (keydb_files[get_keydb_num (key)]);
}

int
kd_env ()
{
  int db_err = 0;

  if (!dbenv)
    {
      if ((db_err = db_env_create (&dbenv, 0)))
	{
	  fprintf (stderr, "Failed allocating database handle (errno = %d)",
		   db_err);
	  exit (db_err);
	}
    }
  return (1);
}

int
kd_add_userid_to_wordlist (llist * wl, unsigned char *userid, long userid_len)
{
  unsigned char *start;
  unsigned char *end;
  words_elem *tmp;
  int ret;

  end = userid;

  while (end < userid + userid_len)
    {
      /* find beginning of word */
      start = end;
      while ((start < userid + userid_len) && !isalnum (*start))
	start++;

      /* find end of word */
      end = start;
      while ((end < userid + userid_len) && isalnum (*end))
	end++;

      /* store it if it's > 1 char */

      if (end - start > 1)
	{
	  if ((tmp = (words_elem *) malloc (sizeof (words_elem))) == NULL)
	    fail ();
	  tmp->ptr = start;
	  tmp->len = end - start;
	  if (!(ret = llist_add_sorted (wl, tmp, words_elem_order)))
	    {
	      free (tmp);
	      fail ();
	    }
	  if (ret == -1)
	    free (tmp);
	}
    }

  return (1);
}

int
sigs_elem_marshall (void *e, void *c)
{
  sigs_elem *se = (sigs_elem *) e;
  xbuffer *xb = (xbuffer *) c;

  return (xbuffer_append (xb, se->sig.buf, se->sig.len));
}

int
userids_elem_marshall (void *e, void *c)
{
  userids_elem *ue = (userids_elem *) e;
  xbuffer *xb = (xbuffer *) c;

  if (!xbuffer_append (xb, ue->uid.buf, ue->uid.len))
    fail ();

  return (llist_iterate (&(ue->sigs), sigs_elem_marshall, c));
}

int
kd_keys_elem_marshall (void *e, void *c)
{
  keys_elem *ke = (keys_elem *) e;
  xbuffer *xb = (xbuffer *) c;

  if (ke->disabled > 0)
    return (1);

  if (!xbuffer_append (xb, ke->pubkey.buf, ke->pubkey.len))
    fail ();

  if (ke->disabled)
    if (!xbuffer_append_str (xb, "\260\001\040"))
      fail ();

  if (!xbuffer_append (xb, ke->revocation.buf, ke->revocation.len))
    fail ();

  if (!xbuffer_append (xb, ke->primary->uid.buf, ke->primary->uid.len))
    fail ();

  if (!llist_iterate (&(ke->primary->sigs), sigs_elem_marshall, c))
    fail ();

  if (!llist_iterate (&(ke->userids), userids_elem_marshall, c))
    fail ();

  if (!xbuffer_append (xb, ke->subkey.buf, ke->subkey.len))
    fail ();

  if (!xbuffer_append (xb, ke->subkeysig.buf, ke->subkeysig.len))
    fail ();

  return 1;
}

int
kd_db_store_keyblock (llist * keys, error * err)
{
  DBT key, newdata;
  DB *db;
  DB_TXN *tid = NULL;
  xbuffer newxb;
  keys_elem *ke = (keys_elem *) (*((void **) keys->xb.buf));

  memset (&key, 0, sizeof (key));
  memset (&newdata, 0, sizeof (newdata));

  /* ke points to first key, which is enough to derive the keyid
     for the database key */

  xbuffer_alloc (&newxb);

  if (!llist_iterate (keys, kd_keys_elem_marshall, (void *) &newxb))
    {
      xbuffer_free (&newxb);
      err->fatal = 1;
      err->str = "internal error while marshalling keyblock";
      fail ();
    }

  key.data = &(ke->keyidbits.buf[4]);
  key.size = KEYDB_KEYID_BYTES;

  newdata.data = (void *) newxb.buf;
  newdata.size = (size_t) newxb.len;

  db = keydb (&key);
  if (use_txn)
    db->dbenv->txn_begin (dbenv, NULL, &tid, 0);
  else
    tid = NULL;

  if (db->put (db, tid, &key, &newdata, 0) < 0)
    {
      if (use_txn)
	tid->abort (tid);
      xbuffer_free (&newxb);
      err->fatal = 1;
      sprintf (err->buf, "error writing to keydb[%d], errno = %d",
	       get_keydb_num (&key), errno);
      fail ();
    }
  if (use_txn)
    TXN_COMMIT (tid, db);
  xbuffer_free (&newxb);

  return (1);
}

/* log utility functions */

void
kd_log_start (char *fct, unsigned char *userid, long len, int flags)
{
  char buf[1024];

  if (userid)
    sprintf (buf, "userid=\"%.*s\"%s, flags=%x",
	     (int) ((len <= 900) ? len : 900), userid,
	     (len <= 900) ? "" : " (truncated)", flags);
  else
    sprintf (buf, "flags=%x", flags);

  log_info (fct, buf);
}

void
kd_log_finish (char *fct, int success)
{
  if (success)
    log_info (fct, "completed successfully");
  else
    log_info (fct, "completed with error");
}

/* create/open/close/sync */

static int
kd_worddata_compare (DB * db, const DBT * a, const DBT * b)
{
  return (memcmp (b->data, a->data, 12));
}

static void
kd_errcall (const char *db_errpfx, char *buffer)
{
  log_error (db_errpfx ? db_errpfx : "db_errcall", buffer);
}

/* if create is non-zero, it specifies the number of database
   files to use.  if it is zero, then the code will read a file
   containing the number of database files to use */
int
kd_open_1 (char *dbdir, int create, int dbtype, int oflags, error * err)
{
  int i, appflags, dbflags, db_err, fd;
  char keydbname[20];
  char line[1024];
  DB_TXN *tid = NULL;
  DBTYPE dbtype2 = DB_UNKNOWN;

  if (dbtype == DB_UNKNOWN)
    {
      if (create != 0)
	{
	  dbtype2 = DB_HASH;
	}
    }
  else if (dbtype == DB_BTREE)
    {
      dbtype2 = DB_BTREE;
    }
  else if (dbtype == DB_HASH)
    {
      dbtype2 = DB_HASH;
    }

  if ((oflags & (KD_OPEN_READONLY | KD_OPEN_RECOVER)) && create)
    {
      err->fatal = 1;
      err->str = "Cannot create or recover database with readonly flag set";
      fail ();
    }

  /* KD_CACHE_SIZE: see config.h */
  dbenv->set_cachesize (dbenv, 0, KD_CACHE_SIZE * 1024 * 1024, 0);
  dbenv->set_errcall (dbenv, kd_errcall);

  appflags = DB_INIT_MPOOL | DB_INIT_LOCK;
  dbflags = 0;
  if (create)
    {
      appflags |= DB_CREATE;
      dbflags |= DB_CREATE;
    }

  if (!(oflags & KD_OPEN_NOTXN) || (oflags & KD_OPEN_RECOVER) || create)
    {
      appflags |= DB_INIT_LOG | DB_INIT_TXN;
      use_txn = 1;
    }
  else
    {
      use_txn = 0;
    }

  if (oflags & KD_OPEN_RECOVER)
    {
      appflags |= DB_INIT_LOG | DB_RECOVER | DB_CREATE;
      dbenv->set_verbose (dbenv, DB_VERB_CHKPOINT, 1);
      dbenv->set_verbose (dbenv, DB_VERB_WAITSFOR, 1);
      dbenv->set_verbose (dbenv, DB_VERB_DEADLOCK, 1);
      dbenv->set_verbose (dbenv, DB_VERB_RECOVERY, 1);
    }

  if ((db_err = dbenv->open (dbenv, dbdir, appflags, 0644)))
    {
      err->fatal = 1;
      sprintf (err->buf, "Error initializing db (errno = %d)", db_err);
      fail ();
    }

  if (create)
    {
      FILE *nkdb;

      sprintf (line, "%s/num_keydb", dbdir);

      if ((nkdb = fopen (line, "r")) == NULL)
	goto do_create;

      if (fgets (line, sizeof (line), nkdb) == NULL)
	{
	  fclose (nkdb);
	  goto do_create;
	}

      if (fclose (nkdb) == EOF)
	goto do_create;

      num_keydb = atoi (line);

      if ((num_keydb < 1) || (num_keydb > 999))
	goto do_create;

      for (i = 0; i < num_keydb; i++)
	{
	  sprintf (line, "%s/keydb%03d", dbdir, i);
	  unlink (line);
	}

    do_create:
      dbflags = DB_CREATE;
      num_keydb = create;

      sprintf (line, "%s/num_keydb", dbdir);

      if ((nkdb = fopen (line, "w")) == NULL)
	{
	  err->fatal = 1;
	  sprintf (err->buf, "Error opening num_keydb (errno = %d)", errno);
	  fail ();
	}

      if (fprintf (nkdb, "%d\n", num_keydb) == 0)
	{
	  fclose (nkdb);
	  err->fatal = 1;
	  sprintf (err->buf, "Error writing to num_keydb (errno = %d)",
		   errno);
	  fail ();
	}

      if (fclose (nkdb) == EOF)
	{
	  err->fatal = 1;
	  sprintf (err->buf, "Error closing num_keydb (errno = %d)", errno);
	  fail ();
	}
    }
  else
    {
      FILE *nkdb;

      if (oflags & KD_OPEN_READONLY)
	dbflags = DB_RDONLY;

      sprintf (line, "%s/num_keydb", dbdir);

      if ((nkdb = fopen (line, "r")) == NULL)
	{
	  err->fatal = 1;
	  sprintf (err->buf, "Error opening num_keydb (errno = %d)", errno);
	  fail ();
	}

      if (fgets (line, sizeof (line), nkdb) == NULL)
	{
	  fclose (nkdb);
	  err->fatal = 1;
	  if (feof (nkdb))
	    sprintf (err->buf, "End of file reading num_keydb");
	  else
	    sprintf (err->buf, "Error reading num_keydb (errno = %d)", errno);
	  fail ();
	}

      if (fclose (nkdb) == EOF)
	{
	  err->fatal = 1;
	  sprintf (err->buf, "Error closing num_keydb (errno = %d)", errno);
	  fail ();
	}

      num_keydb = atoi (line);

      if ((num_keydb < 1) || (num_keydb > 999))
	{
	  err->fatal = 1;
	  sprintf (err->buf, ("num_keydb is out of bounds (%d is not"
			      "between 1 and 999 inclusive)"), errno);
	  fail ();
	}
    }

  if ((keydb_files = (DB **) malloc (sizeof (DB *) * num_keydb)) == NULL)
    {
      err->fatal = 1;
      sprintf (err->buf, "Error allocating keydb state array");
      fail ();
    }

  for (i = 0; i < num_keydb; i++)
    keydb_files[i] = 0;

  for (i = 0; i < num_keydb; i++)
    {
      sprintf (keydbname, "keydb%03d", i);

      sprintf (line, "%s/%s", dbdir, keydbname);
      if (create)
	unlink (line);

      if ((db_err = db_create (&keydb_files[i], dbenv, 0)))
	{
	  err->fatal = 1;
	  sprintf (err->buf, "Error creating database handle %s (errno = %d)",
		   keydbname, db_err);
	  fail ();
	}

      keydb_files[i]->set_pagesize (keydb_files[i], 8192);

      if (use_txn)
	keydb_files[i]->dbenv->txn_begin (keydb_files[i]->dbenv, NULL, &tid,
					  0);
      else
	tid = NULL;

      if ((db_err =
	   keydb_files[i]->open (keydb_files[i], tid, keydbname, NULL,
				 dbtype2, dbflags, 0644)))
	{
	  if (use_txn)
	    tid->abort (tid);

	  err->fatal = 1;
	  sprintf (err->buf, "Error opening %s (errno = %d)", keydbname,
		   db_err);
	  fail ();
	}
      if (use_txn)
	TXN_COMMIT (tid, keydb_files[i]);
      if ((db_err = ((*(keydb_files[i]->fd)) (keydb_files[i], &fd))))
	{
	  err->fatal = 1;
	  sprintf (err->buf, "failed getting %s fd: %d", keydbname, db_err);
	  fail ();
	}

      if (fcntl (fd, F_SETFD, 1) < 0)
	{
	  err->fatal = 1;
	  sprintf (err->buf, "failed making keydb close-on-exec: %d", errno);
	  fail ();
	}
    }

  sprintf (line, "%s/timedb", dbdir);
  if (create)
    unlink (line);

  if ((db_err = db_create (&timedb, dbenv, 0)))
    {
      err->fatal = 1;
      sprintf (err->buf, "Error creating database handle timedb (errno = %d)",
	       db_err);
      fail ();
    }

  if (use_txn)
    timedb->dbenv->txn_begin (timedb->dbenv, NULL, &tid, 0);
  else
    tid = NULL;

  if ((db_err =
       timedb->open (timedb, tid, "timedb", NULL, dbtype2, dbflags, 0644)))
    {
      if (use_txn)
	tid->abort (tid);

      err->fatal = 1;
      sprintf (err->buf, "Error opening timedb (errno = %d)", db_err);
      fail ();
    }
  if (use_txn)
    TXN_COMMIT (tid, timedb);

  if ((db_err = timedb->fd (timedb, &fd)))
    {
      err->fatal = 1;
      sprintf (err->buf, "failed getting timedb fd: %d", db_err);
      fail ();
    }

  if (fcntl (fd, F_SETFD, 1) < 0)
    {
      err->fatal = 1;
      sprintf (err->buf, "failed making timedb close-on-exec: %d", errno);
      fail ();
    }

  sprintf (line, "%s/worddb", dbdir);
  if (create)
    unlink (line);

  if ((db_err = db_create (&worddb, dbenv, 0)))
    {
      err->fatal = 1;
      sprintf (err->buf, "Error creating database handle worddb (errno = %d)",
	       db_err);
      fail ();
    }

  db_err = worddb->set_dup_compare (worddb, kd_worddata_compare);
  db_err = worddb->set_flags (worddb, DB_DUP | DB_DUPSORT);
  db_err = worddb->set_pagesize (worddb, 8192);

  if (use_txn)
    timedb->dbenv->txn_begin (timedb->dbenv, NULL, &tid, 0);
  else
    tid = NULL;

  if ((db_err =
       worddb->open (worddb, tid, "worddb", NULL, dbtype2, dbflags, 0644)))
    {
      if (use_txn)
	tid->abort (tid);

      err->fatal = 1;
      sprintf (err->buf, "Error opening worddb (errno = %d)", db_err);
      fail ();
    }
  if (use_txn)
    TXN_COMMIT (tid, worddb);

  if ((db_err = worddb->fd (worddb, &fd)))
    {
      err->fatal = 1;
      sprintf (err->buf, "failed getting worddb fd: %d", db_err);
      fail ();
    }

  if (fcntl (fd, F_SETFD, 1) < 0)
    {
      err->fatal = 1;
      sprintf (err->buf, "failed making worddb close-on-exec: %d", errno);
      fail ();
    }

  return (1);
}

void
kd_sync_1 ()
{
  int i;

  (*(timedb->sync)) (timedb, 0);
  (*(worddb->sync)) (worddb, 0);
  for (i = 0; i < num_keydb; i++)
    (*(keydb_files[i]->sync)) (keydb_files[i], 0);
}

void
kd_close_1 ()
{
  int count;
  int i;
  int ret = 0;
  char buf[MAXPATHLEN + 100];
  char **loglist, **logfileptr;

  if (timedb)
    timedb->close (timedb, 0);
  if (worddb)
    worddb->close (worddb, 0);
  for (i = 0; i < num_keydb; i++)
    if (keydb_files[i])
      keydb_files[i]->close (keydb_files[i], 0);
  if (dbenv)
    {
      for (count = 0; count < 10; count++)
	{
	  if (use_txn)
	    if ((ret = dbenv->txn_checkpoint (dbenv, 0, 0, 0)))
	      {
		sleep (1);
		continue;
	      }

	  break;
	}

      if (ret)
	{
	  sprintf (buf, "failed db checkpointing: errno = %d", ret);
	  log_error ("kd_close_1", buf);
	  return;
	}

      if (use_txn)
	{
	  if ((ret = dbenv->log_archive (dbenv, &loglist, DB_ARCH_ABS)))
	    {
	      sprintf (buf, "failed listing unneeded log files: errno = %d",
		       ret);
	      log_error ("kd_close_1", buf);
	      return;
	    }

	  if (loglist)
	    {
	      for (logfileptr = loglist; *logfileptr; logfileptr++)
		{
		  if ((ret = unlink (*logfileptr)))
		    {
		      sprintf (buf, "failed removing log file %s: errno = %d",
			       *logfileptr, ret);
		      log_error ("kd_close_1", buf);
		      return;
		    }
		}
	      free (loglist);
	    }
	}

      if ((ret = dbenv->close (dbenv, 0)))
	{
	  sprintf (buf, "failed closing db subsystems: errno = %d", ret);
	  log_error ("kd_close_1", buf);
	}
    }
}

int
kd_create (char *dbdir, int num_files, int dbtype, char **ret)
{
  error err;

  err.str = err.buf;

  kd_env ();
  kd_log_start ("kd_create", NULL, 0, 0);

  if (kd_open_1 (dbdir, num_files, dbtype, 0, &err))
    {
      kd_log_finish ("kd_create", 1);

      return (1);
    }
  else if (!err.fatal)
    {
      if (!(*ret = my_strdup (err.str)))
	{
	  err.fatal = 1;
	  err.str = "Failed allocating space for error string";
	  fail ();

	  /* fall through to fatal error handler */
	}
      else
	{
	  kd_log_finish ("kd_create", 0);

	  return (0);
	}
    }

  /* fatal errors */

  if (err.fatal)
    {
      log_fatal ("kd_open", err.str);
      /* never returns */
    }

  /* keep the compiler quiet */

  return (0);
}

int
kd_open (char *dbdir, int flags, char **ret)
{
  error err;

  err.str = err.buf;

  kd_env ();
  if (kd_open_1 (dbdir, 0, DB_UNKNOWN, flags, &err))
    {
      kd_log_finish ("kd_open", 1);

      return (1);
    }
  else if (!err.fatal)
    {
      if (!(*ret = my_strdup (err.str)))
	{
	  err.fatal = 1;
	  err.str = "Failed allocating space for error string";
	  fail ();

	  /* fall through to fatal error handler */
	}
      else
	{
	  kd_log_finish ("kd_open", 0);

	  return (0);
	}
    }

  /* fatal errors */

  if (err.fatal)
    {
      log_fatal ("kd_open", err.str);
      /* never returns */
    }

  /* keep the compiler quiet */

  return (0);
}

int
kd_sync ()
{
  kd_env ();
  kd_sync_1 ();

  kd_log_finish ("kd_sync", 1);

  return (1);
}

int
kd_close ()
{
  kd_close_1 ();
  kd_log_finish ("kd_close", 1);

  return (1);
}
