/*
 * Portions copyright 2002, 2003 Richard Laager.
 * Copyright (c) 1996, 1997, 1998, 1999, Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARC HOROWITZ ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL MARC HOROWITZ BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

const char rcsid_kd_get_c[] =
  "$Id: kd_get.c,v 1.7 2003/07/31 18:19:02 rlaager Exp $";

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "armor.h"
#include "database.h"
#include "globals.h"
#include "kd_internal.h"
#include "kd_search.h"
#include "kd_types.h"
#include "llist.h"
#include "pgputil.h"

typedef struct _kkec_state
{
  FILE *out;
  xbuffer keys;
}
kkec_state;

int
kd_keys_elem_cout (void *e, void *c)
{
  kkec_state *s = (kkec_state *) c;

  if (!kd_keys_elem_marshall (e, &s->keys))
    fail ();

  fwrite (s->keys.buf, s->keys.len, 1, s->out);

  /* "remove" the data from the buffer, but don't free it.
     this will save a lot of malloc/free calls */
  s->keys.len = 0;

  return (1);
}

int
kd_get_1 (unsigned char *userid, long len, int flags,
	  int maxkeys, ddesc * armored, error * err)
{
  xbuffer keys;
  ddesc binary;

  xbuffer_alloc (&keys);

  /* This is called violating abstractions in the interest of
     efficiency.  Whee. */

  if (flags & KD_GET_STDOUT)
    {
      kkec_state kkecs;
      int ret;

      kkecs.out = stdout;
      xbuffer_alloc (&kkecs.keys);

      ret = kd_search_1 (userid, len, flags & KD_SEARCH_FLAGS, maxkeys,
			 kd_keys_elem_cout, NULL, &kkecs, err);

      xbuffer_free (&kkecs.keys);

      if (!ret)
	return (0);

      armored->data = keys.buf;
      armored->size = keys.len;
      armored->offset = keys.len;

      return (1);
    }

  if (!kd_search_1 (userid, len, flags & KD_SEARCH_FLAGS, maxkeys,
		    kd_keys_elem_marshall, NULL, &keys, err))
    return (0);

  if (keys.len == 0)
    {
      /* this should only happen if the keys requested are disabled
         or the database is corrupt.  If the latter is the case, lie. */

      err->fatal = 0;
      err->str = "The requested key has been disabled";
      return (0);
    }

  if (flags & KD_GET_BINARY)
    {
      armored->data = keys.buf;
      armored->size = keys.len;
      armored->offset = keys.len;
    }
  else
    {
      /* ascii-armor the keyblock */

      binary.data = keys.buf;
      binary.size = keys.len;
      binary.offset = 0;

      armored->size = encode_ascii_size (&binary, "PUBLIC KEY BLOCK");
      if ((armored->data = (unsigned char *) malloc (armored->size)) == NULL)
	{
	  xbuffer_free (&keys);
	  err->fatal = 1;
	  err->str = "Allocating memory for ascii armor key block failed";
	  fail ();
	}
      armored->offset = 0;

      if (!encode_ascii (&binary, "PUBLIC KEY BLOCK", armored))
	{
	  free (armored->data);
	  xbuffer_free (&keys);
	  err->fatal = 1;
	  err->str = "Converting key block to ascii armor failed";
	  fail ();
	}

      xbuffer_free (&keys);
    }

  return (1);
}

int
kd_get (unsigned char *userid, long len, int flags, int maxkeys,
	unsigned char **ret, long *retlen)
{
  error err;
  ddesc armored;

  err.str = err.buf;

  kd_log_start ("kd_get", userid, len, flags);

  if (kd_get_1 (userid, len, flags, maxkeys, &armored, &err))
    {
      *ret = armored.data;
      *retlen = armored.offset;

      kd_log_finish ("kd_get", 1);

      return (1);
    }

  if (!err.fatal)
    {
      if (!(*ret = (unsigned char *) my_strdup (err.str)))
	{
	  err.fatal = 1;
	  err.str = "Failed allocating space for error string";
	  dabort ();

	  /* fall through to fatal error handler */
	}
      else
	{
	  *retlen = strlen ((char *) *ret);

	  kd_log_finish ("kd_get", 0);

	  return (0);
	}
    }

  /* fatal errors */

  if (err.fatal)
    log_fatal ("kd_get", err.str);

  /* keep the compiler quiet */

  return (0);
}
