/*
 * Portions copyright 2002, 2003 Richard Laager.
 * Copyright (c) 1996, 1997, 1998, 1999, Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARC HOROWITZ ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL MARC HOROWITZ BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

const char rcsid_kvcv_c[] =
  "$Id: kvcv.c,v 1.10 2003/07/31 18:45:18 rlaager Exp $";

#include <openssl/md5.h>

#include <openssl/sha.h>

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_TIME_H
#include <time.h>
#endif

#include "database.h"
#include "globals.h"
#include "pgpfile.h"
#include "pgputil.h"

/*
   when you see a type 6 (pubkey), forget all state
   when you see the first type 13, print the pk header
 */

struct state
{
  int got_pk, got_userid, got_revoke;
  ddesc pubkey;
  long pubkeylen;
};

int
kvcv (ddesc * packet, void *c)
{
  long ptype, plen;
  struct state *s = (struct state *) c;

  if (!decode_psf (packet, &ptype, &plen))
    {
      log_error ("decode_psf", "");
      return (0);
    }


  switch (ptype)
    {
    case 6:
      /* public key packet */
      {
	s->got_pk = 1;
	s->got_userid = 0;
	s->got_revoke = 0;
	s->pubkey = *packet;
	s->pubkeylen = plen;
      }
      break;
    case 14:
      /* subkey packet */
      {
	mpidesc m, e;
	unsigned char keytype;
	time_t create_time;
	int keyversion = 4;	/* Assume v4 keys. */

	if (!decode_pubkey (packet, plen, &m, &e, NULL, &keytype,
			    &create_time, &keyversion))
	  {
	    log_error ("decode_pubkey", "");
	    return (0);
	  }

	switch (keytype)
	  {
	  case 1:
	    printf ("subkey (RSA)\n");
	    break;
	  case 16:
	    printf ("subkey (DSA)\n");
	    break;
	  case 17:
	    printf ("subkey (ElGamal)\n");
	    break;
	  default:
	    printf ("subkey (type %d)\n", keytype);
	    break;
	  }
	break;
      }
    case 13:
      /* user id packet */
      {
	int i;
	ddesc userid;

	if (!decode_userid (packet, plen, &userid))
	  {
	    log_error ("decode_userid", "");
	    return (0);
	  }

	if (s->got_userid)
	  {
	    printf ("                              %.*s\n",
		    (int) userid.size, userid.data);
	  }
	else if (s->got_pk)
	  {
	    mpidesc modulus, exponent;
	    unsigned char keytype;
	    time_t create_time;
	    int keyversion = 4;	/* Assume v4 keys. */
	    MD5_CTX md5ctx;
	    SHA_CTX sha;
	    unsigned char hash[20];
	    struct tm *c_tm;

	    if (!decode_pubkey (&(s->pubkey), s->pubkeylen, &modulus,
				&exponent, NULL, &keytype, &create_time,
				&keyversion))
	      {
		log_error ("decode_pubkey 2", "");
		return (0);
	      }

	    if (keytype == 16 || keytype == 17)
	      {
		SHA1_Init (&sha);
		SHA1_Update (&sha, s->pubkey.data, s->pubkey.size);
		SHA1_Final (hash, &sha);
	      }
	    else
	      {
		MD5_Init (&md5ctx);
		MD5_Update (&md5ctx, modulus.number.data,
			    modulus.number.size);
		MD5_Update (&md5ctx, exponent.number.data,
			    exponent.number.size);
		MD5_Final (hash, &md5ctx);
	      }

	    /* pgp does gmtime, so we do, too */
	    c_tm = gmtime (&create_time);

	    printf ("pub%6d/%02X%02X%02X%02X %04d/%02d/%02d %s%.*s\n"
		    "          Key fingerprint =  ",
		    (int) modulus.nbits,
		    modulus.number.data[modulus.number.size - 4],
		    modulus.number.data[modulus.number.size - 3],
		    modulus.number.data[modulus.number.size - 2],
		    modulus.number.data[modulus.number.size - 1],
		    c_tm->tm_year + 1900, c_tm->tm_mon + 1, c_tm->tm_mday,
		    (s->got_revoke ?
		     "*** KEY REVOKED ***\n                              " :
		     ""), (int) userid.size, userid.data);

	    for (i = 0; i < 8; i++)
	      printf ("%02X ", hash[i]);
	    printf (" ");
	    for (i = 8; i < 16; i++)
	      printf ("%02X ", hash[i]);
	    if (keytype == 16 || keytype == 17)
	      {
		printf (" ");
		for (i = 16; i < 20; i++)
		  printf ("%02X ", hash[i]);
	      }

	    printf ("\n");

	    switch (keytype)
	      {
	      case 1:
		printf ("          key type RSA\n");
		break;
	      case 16:
		printf ("          key type DSA\n");
		break;
	      case 17:
		printf ("          key type ElGamal\n");
		break;
	      default:
		printf ("          key type %d\n", keytype);
		break;
	      }

	    s->got_userid = 1;
	  }
      }
      break;
    case 2:
      /* signature packet */
      {
	ddesc keyid;
	/* Point to "null" by default */
	long sigclass;
	time_t sigtime;
	int i;

	keyid.data = NULL;
	if (!decode_sig (packet, plen, &keyid, &sigclass, &sigtime))
	  {
	    log_error ("decode_sig", "");
	    return (0);
	  }

	if (sigclass == 0x20)
	  {
	    /* key revoked */
	    s->got_revoke = 1;
	  }
	else if (sigclass == 0x10)
	  {
	    printf ("sig       ");
	    for (i = 4; i < 8; i++)
	      printf ("%02X", keyid.data[i]);
	    printf ("             (can't do keyid->name conversion yet)\n");
	  }
	else if (sigclass == 0x18)
	  {
	    printf ("             (subkey signature)\n");
	  }
	else
	  {
	    printf ("             (funny signature packet (class %02lX))\n",
		    sigclass);
	  }
      }
      break;
    }

  return (1);
}

int
do_file (ddesc * data)
{
  struct state s;

  s.got_pk = 0;
  s.got_userid = 0;
  s.got_revoke = 0;

  return (decode_file (data, kvcv, (void *) &s));
}
