/*
 * Portions copyright 2002, 2003 Richard Laager.
 * Copyright (c) 1996, 1997, 1998, 1999, Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARC HOROWITZ ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL MARC HOROWITZ BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

const char rcsid_pksclient_c[] =
  "$Id: pksclient.c,v 1.8 2003/03/05 05:05:54 rlaager Exp $";

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#if HAVE_TIME_H
#include <time.h>
#endif

#include "kd_internal.h"
#include "database.h"
#include "globals.h"
#include "util.h"

void
usage (char *argv0)
{
  fprintf (stderr,
	   "%s /db/path create [num_files [db_type]]\n"
	   "%s /db/path recover\n"
	   "%s /db/path add filename [flags]\n"
	   "%s /db/path get userid [flags]\n"
	   "%s /db/path index userid [flags]\n"
	   "%s /db/path since time [flags]\n"
	   "%s /db/path delete userid [flags]\n"
	   "%s /db/path disable userid [flags]\n",
	   argv0, argv0, argv0, argv0, argv0, argv0, argv0, argv0);
  exit (1);
}

int
main (int argc, char *argv[])
{
  char *dir, *op;
  unsigned char *ret;
  long retlen;

  if (argc < 3)
    usage (argv[0]);

  dir = argv[1];
  op = argv[2];

  ret = NULL;

  log_terminal = 1;

  if (strcmp (op, "create") == 0)
    {
      int num_files = 3;
      int dbtype = DB_UNKNOWN;

      if (argv[3])
	{
	  num_files = atoi (argv[3]);
	}

      if (argv[4])
	{
	  if (strcmp (argv[4], "btree") == 0)
	    {
	      dbtype = DB_BTREE;
	    }
	  else if (strcmp (argv[4], "hash") == 0)
	    {
	      dbtype = DB_HASH;
	    }
	  else
	    {
	      fprintf (stderr, "Invalid database type: %s\n", argv[4]);
	    }
	}

      if (!kd_create (dir, num_files, dbtype, (char **) &ret))
	{
	  fprintf (stderr, "database create failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      printf ("Database files created in %s\n", dir);
    }
  else if (strcmp (op, "recover") == 0)
    {
      if (!kd_open (dir, KD_OPEN_RECOVER, (char **) &ret))
	{
	  fprintf (stderr, "database recovery failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      printf ("Database files recovered in %s\n", dir);
    }
  else if (strcmp (op, "add") == 0)
    {
      unsigned char *ret, *newkeys;
      long retlen, newkeyslen;
      char *fn, *opts;
      int donew = 0, flags = 0, oflags = 0;
      FILE *f;
      xfilecontents infile;

      if ((fn = argv[3]) == NULL)
	usage (argv[0]);

      if ((opts = argv[4]))
	{
	  if (strchr (opts, 'n'))
	    donew = 1;
	  if (strchr (opts, 'd'))
	    flags |= KD_ADD_NO_STRIP_DISABLED;
	  if (strchr (opts, 't'))
	    oflags |= KD_OPEN_NOTXN;
	}

      if ((f = fopen (fn, "r")) == NULL)
	{
	  perror ("opening file");
	  exit (1);
	}

      if (!xfilecontents_get (&infile, f))
	{
	  fprintf (stderr, "%s: error getting input file %s contents\n",
		   argv[0], fn);
	  exit (1);
	}

      fclose (f);

      if (!kd_open (dir, oflags, (char **) &ret))
	{
	  fprintf (stderr, "database open failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      if (!kd_add (infile.buf, infile.len, KD_ADD_VERBOSE | flags,
		   &ret, &retlen,
		   donew ? &newkeys : NULL, donew ? &newkeyslen : NULL))
	{
	  fprintf (stderr, "database add failed: %.*s\n", (int) retlen, ret);
	  free (ret);
	  xfilecontents_free (&infile);
	  kd_close ();
	  exit (1);
	}

      xfilecontents_free (&infile);

      fwrite ((void *) ret, (size_t) retlen, 1, stderr);
      if (donew)
	fwrite ((void *) newkeys, (size_t) newkeyslen, 1, stdout);
    }
  else if (strcmp (op, "get") == 0)
    {
      char *userid, *opts;
      int flags = 0, oflags = KD_OPEN_READONLY | KD_OPEN_NOTXN;

      if ((userid = argv[3]) == NULL)
	usage (argv[0]);

      if ((opts = argv[4]))
	{
	  if (strchr (opts, 'e'))
	    flags |= KD_GET_EXACT;
	  if (strchr (opts, 'a'))
	    flags |= KD_GET_ALL;
	  if (strchr (opts, 'b'))
	    flags |= KD_GET_BINARY;
	  if (strchr (opts, 'i'))
	    flags |= KD_GET_IGNORE_ERRORS;
	  if (strchr (opts, 'd'))
	    flags |= KD_GET_RETURN_DISABLED;
	  if (strchr (opts, 's'))
	    flags |= KD_GET_STDOUT;
	  if (strchr (opts, 't'))
	    oflags |= KD_OPEN_NOTXN;
	}

      if (!kd_open (dir, oflags, (char **) &ret))
	{
	  fprintf (stderr, "database open failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      if (!kd_get ((unsigned char *) userid, (long) strlen (userid),
		   flags, -1, &ret, &retlen))
	{
	  fprintf (stderr, "database get failed: %.*s\n", (int) retlen, ret);
	  free (ret);
	  kd_close ();
	  exit (1);
	}

      fwrite ((void *) ret, (size_t) retlen, 1, stdout);
    }
  else if (strcmp (op, "index") == 0)
    {
      char *userid, *opts;
      int flags = 0, oflags = KD_OPEN_READONLY | KD_OPEN_NOTXN;

      if ((userid = argv[3]) == NULL)
	usage (argv[0]);

      if ((opts = argv[4]))
	{
	  if (strchr (opts, 'v'))
	    flags |= KD_INDEX_VERBOSE;
	  if (strchr (opts, 'f'))
	    flags |= KD_INDEX_FINGERPRINT;
	  if (strchr (opts, 'e'))
	    flags |= KD_INDEX_EXACT;
	  if (strchr (opts, 'a'))
	    flags |= KD_INDEX_ALL;
	  if (strchr (opts, 'i'))
	    flags |= KD_INDEX_IGNORE_ERRORS;
	  if (strchr (opts, 's'))
	    flags |= KD_INDEX_STDOUT;
	  if (strchr (opts, 't'))
	    oflags |= KD_OPEN_NOTXN;
	}

      if (!kd_open (dir, oflags, (char **) &ret))
	{
	  fprintf (stderr, "database open failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      if (!kd_index ((unsigned char *) userid, (long) strlen (userid),
		     flags, -1, &ret, &retlen))
	{
	  fprintf (stderr, "database index failed: %.*s\n", (int) retlen,
		   ret);
	  free (ret);
	  kd_close ();
	  exit (1);
	}

      fwrite ((void *) ret, (size_t) retlen, 1, stdout);
    }
  else if (strcmp (op, "since") == 0)
    {
      time_t since, last;
      char *opts;
      int flags = 0, oflags = KD_OPEN_READONLY | KD_OPEN_NOTXN;

      if (argv[3] == NULL)
	usage (argv[0]);

      since = atoi (argv[3]);

      if ((opts = argv[4]))
	{
	  if (strchr (opts, 'b'))
	    flags |= KD_SINCE_BINARY;
	  if (strchr (opts, 'r'))
	    {
	      time_t now;
	      now = time (NULL);
	      since = now - since;
	    }
	  if (strchr (opts, 't'))
	    oflags |= KD_OPEN_NOTXN;
	}

      if (!kd_open (dir, oflags, (char **) &ret))
	{
	  fprintf (stderr, "database open failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      if (!kd_since (since, flags, -1, &last, &ret, &retlen))
	{
	  fprintf (stderr, "database since failed: %.*s\n", (int) retlen,
		   ret);
	  free (ret);
	  kd_close ();
	  exit (1);
	}

      fprintf (stderr, "Last key added at %ld\n\n", (long) last);

      fwrite ((void *) ret, (size_t) retlen, 1, stdout);
    }
  else if (strcmp (op, "delete") == 0)
    {
      char *userid, *opts;
      int oflags = 0;

      if ((userid = argv[3]) == NULL)
	usage (argv[0]);

      if ((opts = argv[4]))
	{
	  if (strchr (opts, 't'))
	    oflags |= KD_OPEN_NOTXN;
	}

      if (!kd_open (dir, oflags, (char **) &ret))
	{
	  fprintf (stderr, "database open failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      if (!kd_delete ((unsigned char *) userid, (long) strlen (userid),
		      0, &ret, &retlen))
	{
	  fprintf (stderr, "database delete failed: %.*s\n", (int) retlen,
		   ret);
	  free (ret);
	  kd_close ();
	  exit (1);
	}

      fwrite ((void *) ret, (size_t) retlen, 1, stdout);
    }
  else if (strcmp (op, "disable") == 0)
    {
      char *userid, *opts;
      int flags = 0, oflags = 0;

      if ((userid = argv[3]) == NULL)
	usage (argv[0]);

      if ((opts = argv[4]))
	{
	  if (strchr (opts, 'c'))
	    flags |= KD_DISABLE_CLEAR;
	  if (strchr (opts, 't'))
	    oflags |= KD_OPEN_NOTXN;
	}

      if (!kd_open (dir, oflags, (char **) &ret))
	{
	  fprintf (stderr, "database open failed: %s\n", ret);
	  free (ret);
	  exit (1);
	}

      if (!kd_disable ((unsigned char *) userid, (long) strlen (userid),
		       flags, &ret, &retlen))
	{
	  fprintf (stderr, "database disable failed: %.*s\n", (int) retlen,
		   ret);
	  free (ret);
	  kd_close ();
	  exit (1);
	}

      fwrite ((void *) ret, (size_t) retlen, 1, stdout);
    }
  else
    {
      usage (argv[0]);
    }

  if (ret)
    free (ret);

  kd_close ();

  exit (0);
}
