/*
 * Portions copyright 2002, 2003 Richard Laager.
 * Copyright (c) 1996, 1997, 1998, 1999, Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARC HOROWITZ ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL MARC HOROWITZ BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

const char rcsid_pksd_c[] =
  "$Id: pksd.c,v 1.19 2003/04/27 02:46:45 rlaager Exp $";

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if HAVE_ERRNO_H
#include <errno.h>
#endif

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_SIGNAL_H
#include <signal.h>
#endif

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#if HAVE_SYSLOG_H
#include <syslog.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "database.h"
#include "globals.h"
#include "mail_req.h"
#include "mail_send.h"
#include "multiplex.h"
#include "pks_config.h"
#include "pks_incr.h"
#include "pks_socket.h"
#include "pks_www.h"
#include "util.h"

extern int errno;
extern char *argv0;

char *write_pidfile (char *);

int
main (int argc, char *argv[])
{
  char *ret;
  pks_config pc;
  mail_send_conf msc;
  pks_incr_conf pic;
  mail_req_conf mrc;
  pks_socket_conf psc;
  pks_www_conf pwc;
  int retval;
  char *x, *y;
  char *pidfile = NULL;
  char *pid_dir = NULL;

  /* just in case we get called with a full pathname */
  x = argv[0];
  while ((y = strtok (x, "/")))
    {
      argv0 = y;
      x = NULL;
    }
  argv[0] = argv0;

#ifdef LOG_LOCAL2
  openlog (argv0, LOG_PID, LOG_LOCAL2);
#else
  /* bsd4.2/ultrix */
  openlog (argv0, LOG_PID);
#endif

#if DEBUG
  if (argc > 1)
    {
      if (strcmp (argv[1], "-D") == 0)
	{
	  debug = 1;
	  argc--;
	  argv++;
	}
    }
#endif

  if (argc != 2)
    {
      char buf[1024];

      sprintf (buf, "usage: %s conf_file", argv0);
      fprintf (stderr, "%s\n", buf);
      log_fatal ("main", buf);
    }

  if (!pks_read_conf_file (argv[1], &pc))
    {
      fprintf (stderr, "error reading configuration file\n"
	       "Did you forget to configure the server?\n"
	       "See the INSTALL file for help.\n");
      log_fatal ("main", "error reading configuration file");
    }

  /* the proper way to use chroot:  chroot(foo) and then chdir(/) */
  if (pc.chroot_dir[0] == '/')
    {
      if (getuid () != 0)
	{
	  fprintf (stderr, "chroot(%s) requested - permission denied\n",
		   pc.chroot_dir);
	  log_error ("main",
		     "chroot requested, but only root can call chroot");
	}
      else
	{
	  errno = 0;
	  if (chroot (pc.chroot_dir) || chdir ("/"))
	    {
	      fprintf (stderr, "chroot(%s) failed - errno %d\n",
		       pc.chroot_dir, errno);
	      log_fatal ("main", "chroot failed");
	    }
	}
    }

  msc.mail_delivery_client = pc.mail_delivery_client;
  msc.maintainer_email = pc.maintainer_email;
  msc.mail_intro_file = pc.mail_intro_file;
  msc.help_dir = pc.help_dir;
  msc.default_lang = pc.default_lang;

  pic.this_site = pc.this_site;
  pic.syncsites = &(pc.syncsites);
  pic.msc = &msc;

  mrc.msc = &msc;
  mrc.pic = &pic;
  mrc.max_last = pc.max_last;
  mrc.max_last_reply_keys = pc.max_last_reply_keys;
  mrc.max_reply_keys = pc.max_reply_keys;

  psc.socket = pc.socket_name;
  psc.mrc = &mrc;

  pwc.addrs = &(pc.www_addrs);
  pwc.port = pc.www_port;
  pwc.readonly = pc.www_readonly;
  pwc.www_dir = pc.www_dir;
  pwc.max_reply_keys = pc.max_reply_keys;
  pwc.pic = &pic;
  pid_dir = pc.pid_dir;

  if (pc.pksd_uid != 0)
    setuid (pc.pksd_uid);
  if (pc.pksd_gid != 0)
    setgid (pc.pksd_gid);

  if (!pc.www_port && (pc.mail_delivery_client[0] == '\0'))
    {
      fprintf (stderr, "The configuration file did not specify any servers\n"
	       "to run.  Did you forget to configure the server?\n"
	       "See the INSTALL file for help.\n");
      log_fatal ("main", "nothing to do!");
      exit (1);
    }

  if (!kd_install_sig_handler (&kd_sig_handler))
    log_fatal ("main", "installing kd_sig_handler");

  if (!mp_ignore_signals ())
    log_fatal ("main", "ignoring multiplexor signals");

  /* XXX there is no way to run the daemon without transactions.
     I do not consider this a major problem --marc */
  if (!kd_open (pc.db_dir, 0, &ret))
    log_fatal ("main", "database open failed");

  if (!mp_init ())
    log_fatal ("main", "failed initializing multiplexor");

  if (pc.www_port)
    pks_www_init (&pwc);

  if (pc.mail_delivery_client[0])
    pks_socket_init (&psc);

  log_info ("main", "pks server started");
  if (pid_dir)
    pidfile = write_pidfile (pid_dir);

  retval = mp_go ();

  kd_close ();

  switch (retval)
    {
    case MP_NODESC:
      log_error ("main", "last multiplexer descriptor deleted unexpectedly");
      break;
    case MP_ERROR:
      log_fatal ("main", "error in multiplexor main loop");
      break;
    case EXIT_SIGNAL:
      log_info ("main", "pksd terminated by signal");
      break;

      /* case EXIT_SHUTDOWN: pks server shut down by pksdctl */
    }


  if (pidfile)
    unlink (pidfile);
  log_info ("main", "pks server exited");

  exit (0);
}

char *
write_pidfile (char *dir)
{
  long int fd, oldpid;
  FILE *fp;
  char pidfile[1024], tempfile[1024];

  if (!dir)
    return (NULL);

  errno = 0;
  snprintf (pidfile, sizeof (pidfile) - 1, "%s/%s.pid", dir, argv0);
  if ((fd = open (pidfile, O_EXCL | O_CREAT | O_WRONLY, 0644)) != -1)
    {
      /* yay! no pre-existing pid file! */
      fp = fdopen (fd, "w");
      if (errno == 0)
	{
	  fprintf (fp, "%ld", (long) getpid ());
	  fclose (fp);
	  close (fd);
	  return (my_strdup (pidfile));
	}
      close (fd);
      unlink (pidfile);
    }
  else
    {
      errno = 0;
      fp = fopen (pidfile, "r");
      if (errno == 0)
	{
	  fscanf (fp, "%ld", &oldpid);
	  fclose (fp);
	  kill (oldpid, 0);
	  if ((errno == ESRCH) || (errno == EPERM))
	    {
	      /* oldpid doesn't exist, or stale pid file */
	      snprintf (tempfile, sizeof (tempfile) - 1, "%s/pid.XXXXXXXXXX",
			dir);
	      errno = 0;
	      fd = mkstemp (tempfile);
	      if (errno == 0)
		{
		  fp = fdopen (fd, "w");
		  if (errno == 0)
		    {
		      fprintf (fp, "%ld", (long) getpid ());
		      fclose (fp);
		      close (fd);
		      rename (tempfile, pidfile);
		      if (errno == 0)
			{
			  return (strdup (pidfile));
			}
		    }
		}
	      close (fd);
	      unlink (tempfile);
	    }
	}
    }
  /* oldpid is running, and as my userid. assume the worst */
  log_info ("write_pidfile", "unable to write pidfile");
  fprintf (stderr, "Arrgh! Failed to write a pidfile!\n");
  return (NULL);
}
